from unittest import mock

from django.test.testcases import TestCase


class TestDecoratorTestCase(TestCase):

    def tearDown(self):
        from post_deploy.utils import register_post_deploy
        register_post_deploy.bindings.clear()

    @mock.patch('post_deploy.utils.register_post_deploy.register_input')
    def test_plain_decorator(self, mocked_register_input):
        from post_deploy.utils import register_post_deploy

        @register_post_deploy
        def test_method_1():
            pass

        registered_names = [arg.__name__ for arg in mocked_register_input.mock_calls[0].args]
        self.assertIn('test_method_1', registered_names)

    @mock.patch('post_deploy.utils.register_post_deploy.register_input')
    def test_noargs_decorator(self, mocked_register_input):
        from post_deploy.utils import register_post_deploy

        @register_post_deploy()
        def test_method_2():
            pass

        registered_names = [arg.__name__ for arg in mocked_register_input.mock_calls[0].args]
        self.assertIn('test_method_2', registered_names)

    def test_auto_false_decorator(self):
        from post_deploy.utils import register_post_deploy

        @register_post_deploy(auto=False)
        def test_method_3():
            pass

        self.assertIn('post_deploy.test_method_3', register_post_deploy.bindings)
        self.assertFalse(register_post_deploy.bindings['post_deploy.test_method_3']['auto'])

    def test_auto_true_when_empty_decorator(self):
        from post_deploy.utils import register_post_deploy

        @register_post_deploy
        def test_method_4():
            pass

        @register_post_deploy()
        def test_method_5():
            pass

        @register_post_deploy()
        def test_method_6(auto=True):
            pass

        self.assertIn('post_deploy.test_method_4', register_post_deploy.bindings)
        self.assertTrue(register_post_deploy.bindings['post_deploy.test_method_4']['auto'])

        self.assertIn('post_deploy.test_method_5', register_post_deploy.bindings)
        self.assertTrue(register_post_deploy.bindings['post_deploy.test_method_5']['auto'])

        self.assertIn('post_deploy.test_method_6', register_post_deploy.bindings)
        self.assertTrue(register_post_deploy.bindings['post_deploy.test_method_6']['auto'])



from contextlib import contextmanager


class DefaultContext():

    def __init__(self, parameters):
        self.parameters = parameters or self.default_parameters()

    def default_parameters(self):
        return {}

    @contextmanager
    def execute(self):
        yield





import uuid


class DefaultScheduler():

    def task_ready(self, id):
        return False

    def schedule(self, action_ids, context_kwargs):
        from post_deploy.local_utils import run_deploy_action
        run_deploy_action(action_ids)
        return uuid.uuid4()



from inspect import isfunction

from django.conf import settings
from django.utils import timezone
from django.utils.module_loading import import_string

from post_deploy.models import PostDeployAction
from post_deploy.plugins.context import DefaultContext
from post_deploy.plugins.scheduler import DefaultScheduler
from post_deploy.utils import register_post_deploy


def initialize_actions():
    from post_deploy.utils import register_post_deploy
    action_list = {}

    for app in settings.INSTALLED_APPS:
        try:
            __import__(f"{app}.post_deploy")
        except ModuleNotFoundError:
            pass

    for id, configuration in register_post_deploy.bindings.items():
        action, created = PostDeployAction.objects.get_or_create(id=id)
        action.auto = configuration['auto']
        action.description = configuration['description']
        action.sync_status()
        action.save()

        action_list[id] = action

    current_actions = action_list.keys()
    for action in PostDeployAction.objects.filter(available=True):
        if action.id not in current_actions:
            action.available = False
            action.save()

    return action_list


def run_deploy_action(action_uuids):
    initialize_actions()
    for uuid in action_uuids:
        action = PostDeployAction.objects.get(uuid=uuid)
        config = register_post_deploy.bindings.get(action.id)

        if config and action:
            try:
                vehicle = config['class']
                if isfunction(vehicle):
                    vehicle()
                else:
                    raise Exception(f"{vehicle} is not a function")
            except Exception as e:
                action.message = str(e)
            finally:
                action.completed_at = timezone.localtime()
                action.done = True
                action.save()


def get_context_manager(context_parameters) -> DefaultContext:
    Manager = import_string(
        getattr(settings, 'POST_DEPLOY_CONTEXT_MANAGER', 'post_deploy.plugins.context.DefaultContext'))
    return Manager(context_parameters)


def get_scheduler_manager() -> DefaultScheduler:
    Manager = import_string(
        getattr(settings, 'POST_DEPLOY_SCHEDULER_MANAGER', 'post_deploy.plugins.scheduler.DefaultScheduler'))
    return Manager()



import uuid

from django.db import models
from django.utils import timezone


class PostDeployActionQueryset(models.QuerySet):

    def available(self):
        return self.filter(available=True)

    def unprocessed(self):
        return self.available().filter(started_at__isnull=True)

    def todo(self):
        return self.unprocessed().filter(auto=True)

    def manual(self):
        return self.unprocessed().filter(auto=False)

    def completed(self):
        return self.filter(done=True, message__isnull=True)

    def with_errors(self):
        return self.filter(message__isnull=False)

    def running(self):
        return self.filter(started_at__isnull=False, done=False)

    def ids(self):
        return [id for id in self.values_list('uuid', flat=True)]


class PostDeployAction(models.Model):
    objects = PostDeployActionQueryset.as_manager()

    id = models.CharField(max_length=255, primary_key=True)
    uuid = models.UUIDField(default=uuid.uuid4, null=False)
    created_at = models.DateTimeField(auto_now_add=True)

    started_at = models.DateTimeField(blank=True, null=True)
    completed_at = models.DateTimeField(blank=True, null=True)

    auto = models.BooleanField(default=True)
    description = models.TextField(blank=True, null=True)

    available = models.BooleanField(default=True)
    done = models.BooleanField(default=False)

    task_id = models.UUIDField(blank=True, null=True)
    message = models.TextField(blank=True, null=True)

    def __str__(self):
        return self.id

    def sync_status(self):
        if not self.task_id or self.done:
            return

        from post_deploy.local_utils import get_scheduler_manager
        scheduler = get_scheduler_manager()

        if scheduler.task_ready(str(self.task_id)):
            self.done = True
            self.completed_at = timezone.localtime()
            self.message = "Abnormal termination detected. Please check the logs for details."



from django.core.management.base import BaseCommand
from django.utils import timezone

from post_deploy.models import PostDeployAction
from post_deploy.local_utils import initialize_actions, get_context_manager, get_scheduler_manager


class Command(BaseCommand):
    help = "Execute post deployment actions."

    def __init__(self):
        super(Command, self).__init__()
        self.context_manager = None

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser=parser)
        parser.add_argument('--report', const=True, action='store_const',
                            help="Print the status of all actions.")
        parser.add_argument('--auto', const=True, action='store_const',
                            help="Execute all pending actions that have auto=True (default setting).")
        parser.add_argument('--all', const=True, action='store_const',
                            help="Execute all pending actions no matter the value of auto.")
        parser.add_argument('--one', help="Execute one of the actions.")

    def handle(self, *args, **options):
        self.context_manager = get_context_manager(None)
        with self.context_manager.execute():
            todo_list = []
            for todo in ['report', 'auto', 'all', 'one']:
                if options.get(todo):
                    todo_list.append(todo)

            if len(todo_list) != 1:
                self.stderr.write("Provide 1 todo at a time.\n")
                return self.do_help()

            initialize_actions()

            if options['report']:
                return self.do_report()

            if PostDeployAction.objects.running().exists():
                self.stderr.write("Please wait until all tasks are completed.")
                return

            if options['auto']:
                return self.do_execute(PostDeployAction.objects.todo().ids())

            if options['all']:
                return self.do_execute(PostDeployAction.objects.unprocessed().ids())

            if options['one']:
                return self.do_execute([options['one']])

    def do_help(self):
        self.print_help("manage.py", "post_deploy")

    def do_report(self):
        if PostDeployAction.objects.unprocessed().count() == 0:
            self.stdout.write("No pending actions found.")

        if PostDeployAction.objects.todo().exists():
            self.stdout.write("Pending actions that can run automatically:")
            for action in PostDeployAction.objects.todo():
                self.stdout.write("* %s" % action.id)

        if PostDeployAction.objects.manual().exists():
            self.stdout.write("\nPending actions that need starting manually:")
            for action in PostDeployAction.objects.manual():
                if action.message:
                    self.stdout.write(f"* {action.id} ({action.message})")
                else:
                    self.stdout.write(f"* {action.id}")

        if PostDeployAction.objects.running().exists():
            self.stdout.write("\nCurrently running actions:")
            for action in PostDeployAction.objects.running():
                self.stdout.write(f"* {action.id} ({action.started_at})")

        if PostDeployAction.objects.with_errors().exists():
            self.stdout.write("\nActions that failed:")
            for action in PostDeployAction.objects.with_errors():
                self.stdout.write(f"* {action.id} ({action.completed_at})")
                self.stdout.write(f"  {action.message}")

        if PostDeployAction.objects.completed().order_by('-completed_at').exists():
            self.stdout.write("\nCompleted actions:")
            for action in PostDeployAction.objects.completed().order_by('-completed_at'):
                self.stdout.write(f"* {action.id} ({action.completed_at})")

    def do_execute(self, ids):
        actions = [action for action in PostDeployAction.objects.filter(uuid__in=ids)]
        if len(actions) == 0:
            self.stderr.write(f"No tasks available.")
            return

        real_ids = [action.uuid for action in actions]
        PostDeployAction.objects.filter(uuid__in=real_ids).update(
            started_at=timezone.localtime(),
            completed_at=None,
            message=None,
            done=False
        )

        self.stdout.write("Scheduled execute:")
        for action in actions:
            self.stdout.write(f"* {action.id}")

        task_id = get_scheduler_manager().schedule(real_ids, self.context_manager.default_parameters())
        PostDeployAction.objects.filter(uuid__in=real_ids).update(
            task_id=task_id
        )



from celery import Celery
from celery.result import AsyncResult

from django.conf import settings
from django.utils.module_loading import import_string

from post_deploy.errors import PostDeployConfigurationError
from post_deploy.plugins.scheduler import DefaultScheduler


class CeleryScheduler(DefaultScheduler):

    @staticmethod
    def get_celery_app():
        try:
            app = import_string(settings.POST_DEPLOY_CELERY_APP)
            if not isinstance(app, Celery):
                raise PostDeployConfigurationError(
                    f"{settings.POST_DEPLOY_CELERY_APP or 'None'} is not a valid Celery app.")
            return app
        except AttributeError:
            raise PostDeployConfigurationError(
                "Put the path to the celery app in POST_DEPLOY_CELERY at your project settings.")

    def task_ready(self, id):
        app = CeleryScheduler.get_celery_app()
        return AsyncResult(id=id, app=app).ready()

    def schedule(self, action_ids, context_kwargs):
        from post_deploy.tasks import deploy_task
        result = deploy_task.delay(action_ids, context_kwargs)
        return result.id



from contextlib import contextmanager
from django_tenants.utils import parse_tenant_config_path, schema_context

from post_deploy.plugins.context import DefaultContext


class TenantContext(DefaultContext):

    def default_parameters(self):
        return {
            'schema': parse_tenant_config_path("")
        }

    @contextmanager
    def execute(self):
        with schema_context(self.parameters['schema']):
            yield



from post_deploy.local_utils import get_context_manager, run_deploy_action
from post_deploy.plugins.scheduler.celery import CeleryScheduler

app = CeleryScheduler.get_celery_app()


@app.task
def deploy_task(task_ids, context_parameters):
    manager = get_context_manager(context_parameters)
    with manager.execute():
        run_deploy_action(task_ids)



import io
from distutils.core import setup
from os.path import exists

from setuptools import find_packages
from post_deploy import VERSION

setup(
    name='django_post_deploy',
    packages=find_packages(),
    version=VERSION[1:],
    license='cc-by-4.0',
    description='A generic way to have post-deploy actions done.',
    long_description=io.open("README.md", encoding='utf-8').read() if exists("README.md") else "",
    long_description_content_type='text/markdown',
    author='Erlend ter Maat',
    author_email='erwitema@gmail.com',
    url='https://github.com/erlendgit/django_post_deploy',
    download_url=f'https://github.com/erlendgit/django_post_deploy/archive/refs/tags/{VERSION}.tar.gz',
    keywords=['Django', 'Deployment', 'Management', 'CLI'],
    install_requires=[
        'Django',
    ],
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',
    ],
)



# Generated by Django 4.0.4 on 2022-05-20 21:04

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='PostDeployAction',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('uuid', models.UUIDField(default=uuid.uuid4)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('started_at', models.DateTimeField(blank=True, null=True)),
                ('completed_at', models.DateTimeField(blank=True, null=True)),
                ('auto', models.BooleanField(default=True)),
                ('description', models.TextField(blank=True, null=True)),
                ('available', models.BooleanField(default=True)),
                ('done', models.BooleanField(default=False)),
                ('task_id', models.UUIDField(blank=True, null=True)),
                ('message', models.TextField(blank=True, null=True)),
            ],
        ),
    ]



from collections import OrderedDict
from functools import update_wrapper, partial
from inspect import isfunction


class register_post_deploy():
    bindings = OrderedDict()

    decorated = None

    def __init__(self, *args, **kwargs):
        # Copy configuration. Provide defaults if kwargs is empty.
        self.auto = kwargs.get('auto', True)
        self.description = kwargs.get('description', "")

        if self.args_has_one_function(args):
            # Decorated without configuration.
            self.register_input(*args)
            self.decorated = args[0]
            update_wrapper(self, self.decorated)

    def args_has_one_function(self, args):
        if len(args) == 0:
            return False
        if len(args) > 1 or not isfunction(args[0]):
            raise ValueError("Specify keyword arguments when using this decorator.")
        return True

    def __get__(self, obj):
        return partial(self, obj)

    def __call__(self, *args, **kwargs):
        if not self.decorated:
            # Decorated with configuration.
            self.register_input(*args)
            return args[0]

        return self.decorated(*args, **kwargs)

    def register_input(self, c):
        app = c.__module__.split('.')[0]
        key = f"{app}.{c.__name__}"
        register_post_deploy.bindings[key] = {
            "class": c,
            "auto": self.auto,
            "description": self.description,
        }
